package atomicstryker.multimine.client;

import atomicstryker.multimine.common.MultiMine;
import atomicstryker.multimine.common.PartiallyMinedBlock;
import atomicstryker.multimine.common.network.PartialBlockPacket;
import net.minecraft.block.Block;
import net.minecraft.block.SoundType;
import net.minecraft.block.state.IBlockState;
import net.minecraft.client.Minecraft;
import net.minecraft.client.audio.PositionedSoundRecord;
import net.minecraft.client.renderer.DestroyBlockProgress;
import net.minecraft.client.renderer.RenderGlobal;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.item.ItemStack;
import net.minecraft.util.EnumHand;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.fml.client.FMLClientHandler;

import java.lang.reflect.Field;
import java.util.Map;

public class MultiMineClient
{
    private static MultiMineClient instance;
    private static Minecraft mc;
    private static EntityPlayer thePlayer;
    private final PartiallyMinedBlock[] partiallyMinedBlocksArray;
    private Map<Integer, DestroyBlockProgress> vanillaDestroyBlockProgressMap;
    private int arrayOverWriteIndex;
    private BlockPos curBlock;
    private float lastBlockCompletion;
    private int lastCloudTickReading;

    /**
     * Client instance of Multi Mine Mod. Keeps track of whether or not the current Server has the Mod,
     * the current Block being mined, and hacks into the vanilla "partially Destroyed Blocks" RenderMap.
     * Also handles Packets sent from server to announce other people's damaged Blocks.
     */
    @SuppressWarnings({ "unchecked", "rawtypes" })
    public MultiMineClient()
    {
        instance = this;
        mc = FMLClientHandler.instance().getClient();
        partiallyMinedBlocksArray = new PartiallyMinedBlock[30];
        arrayOverWriteIndex = 0;
        curBlock = BlockPos.field_177992_a;
        lastBlockCompletion = 0F;
        lastCloudTickReading = 0;

        MultiMine.instance().debugPrint("Multi Mine about to hack vanilla RenderMap");
        for (Field f : RenderGlobal.class.getDeclaredFields())
        {
            if (f.getType().equals(Map.class))
            {
                f.setAccessible(true);
                try
                {
                    vanillaDestroyBlockProgressMap = (Map) f.get(mc.field_71438_f);
                    MultiMine.instance().debugPrint("Multi Mine vanilla RenderMap invasion successful, field: " + f.getName());
                    break;
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        }
    }

    public static MultiMineClient instance()
    {
        return instance;
    }

    /**
     * Called by the transformed PlayerControllerMP, has the ability to override the internal block mine completion with another value
     *
     * @param pos             BlockPos instance being hit
     * @param blockCompletion mine completion value as currently held by the controller. Values >= 1.0f trigger block breaking
     * @return value to override blockCompletion in PlayerControllerMP with
     */
    @SuppressWarnings("unused")
    public float eventPlayerDamageBlock(BlockPos pos, float blockCompletion)
    {
        thePlayer = FMLClientHandler.instance().getClient().field_71439_g;
        boolean cachedProgressWasAhead = false;
        // see if we have multimine completion cached somewhere
        for (int i = 0; i < partiallyMinedBlocksArray.length; i++)
        {
            if (partiallyMinedBlocksArray[i] != null && partiallyMinedBlocksArray[i].getPos().equals(pos))
            {
                float savedProgress = partiallyMinedBlocksArray[i].getProgress();
                MultiMine.instance().debugPrint("found cached block at index " + i + ", cached: " + savedProgress + ", completion: " + blockCompletion);
                if (savedProgress > blockCompletion)
                {
                    lastBlockCompletion = savedProgress;
                    cachedProgressWasAhead = true;
                }
                break;
            }
        }

        if (!cachedProgressWasAhead)
        {
            if (!curBlock.equals(pos))
            {
                // setup new block values
                curBlock = pos;
                lastBlockCompletion = blockCompletion;
            }
            else if (blockCompletion > lastBlockCompletion)
            {
                MultiMine.instance().debugPrint("Client has block progress for: [" + pos + "], actual completion: " + blockCompletion + ", lastCompletion: " + lastBlockCompletion);
                MultiMine.instance().networkHelper.sendPacketToServer(new PartialBlockPacket(thePlayer.func_70005_c_(), curBlock.func_177958_n(), curBlock.func_177956_o(), curBlock.func_177952_p(), blockCompletion));
                MultiMine.instance().debugPrint("Sent block progress packet to server: " + blockCompletion);
                lastBlockCompletion = blockCompletion;
                updateLocalPartialBlock(curBlock.func_177958_n(), curBlock.func_177956_o(), curBlock.func_177952_p(), blockCompletion);
            }
        }

        return lastBlockCompletion;
    }

    /**
     * Helper method to emulate the Digging Particles created when a player mines a Block. This usually runs every tick while mining
     *
     * @param x coordinate of Block being mined
     * @param y coordinate of Block being mined
     * @param z coordinate of Block being mined
     */
    private void renderBlockDigParticles(int x, int y, int z)
    {
        World world = thePlayer.field_70170_p;
        IBlockState state = world.func_180495_p(new BlockPos(x, y, z));
        Block block = state.func_177230_c();
        if (block != Blocks.field_150350_a)
        {
            SoundType soundtype = block.func_185467_w();
            mc.func_147118_V()
                    .func_147682_a(new PositionedSoundRecord(soundtype.func_185846_f(), SoundCategory.NEUTRAL, (soundtype.func_185843_a() + 1.0F) / 8.0F, soundtype.func_185847_b() * 0.5F, new BlockPos(x, y, z)));
            mc.field_71452_i.func_180533_a(new BlockPos(x, y, z), state);
        }
    }

    /**
     * Called when a server informs the client about new Block progress. See if it exists locally and update, else add it.
     *
     * @param x        coordinate of Block
     * @param y        coordinate of Block
     * @param z        coordinate of Block
     * @param progress of Block Mining, float 0 to 1
     */
    public void onServerSentPartialBlockData(int x, int y, int z, float progress)
    {
        if (thePlayer == null)
        {
            return;
        }

        MultiMine.instance().debugPrint("Client received partial Block packet for: [" + x + "|" + y + "|" + z + "], progress now: " + progress);
        updateLocalPartialBlock(x, y, z, progress);
    }

    private void updateLocalPartialBlock(int x, int y, int z, float progress)
    {
        updateCloudTickReading();

        EntityPlayer player = thePlayer;
        World w = player.field_70170_p;
        BlockPos pos = new BlockPos(x, y, z);
        final IBlockState block = w.func_180495_p(pos);

        final PartiallyMinedBlock newBlock = new PartiallyMinedBlock(x, y, z, thePlayer.field_71093_bK, progress);
        PartiallyMinedBlock iterBlock;
        int freeIndex = -1;
        for (int i = 0; i < partiallyMinedBlocksArray.length; i++)
        {
            iterBlock = partiallyMinedBlocksArray[i];
            if (iterBlock == null && freeIndex == -1)
            {
                freeIndex = i;
            }
            else if (newBlock.equals(iterBlock))
            {
                boolean notClientsBlock = false;
                // if other guy's progress advances, render digging
                if (iterBlock.getProgress() < progress && !iterBlock.getPos().equals(pos))
                {
                    renderBlockDigParticles(x, y, z);
                    notClientsBlock = true;
                }
                MultiMine.instance().debugPrint(
                        "Client updating local partial block [" + x + "|" + y + "|" + z + "], at index " + i + ", notClientsBlock: " + notClientsBlock + ", setting progres from " + iterBlock
                                .getProgress() + " to " + progress);

                iterBlock.setProgress(progress);
                final DestroyBlockProgress newDestroyBP = new DestroyBlockProgress(0, iterBlock.getPos());
                newDestroyBP.func_73107_a(Math.min(9, Math.round(10f * iterBlock.getProgress())));
                newDestroyBP.func_82744_b(lastCloudTickReading);
                vanillaDestroyBlockProgressMap.put(i, newDestroyBP);

                if (iterBlock.isFinished())
                {
                    w.func_175715_c(player.func_145782_y(), pos, -1);

                    if (block.func_177230_c() != Blocks.field_150350_a)
                    {
                        if (!notClientsBlock && block.func_177230_c().removedByPlayer(w.func_180495_p(pos), w, pos, player, true))
                        {
                            block.func_177230_c().func_176206_d(w, pos, block);
                            block.func_177230_c().func_180657_a(w, player, pos, block, w.func_175625_s(pos), player.func_184614_ca());
                        }

                        w.func_184133_a(null, pos, block.func_177230_c().func_185467_w().func_185845_c(), SoundCategory.BLOCKS, block.func_177230_c().func_185467_w().func_185843_a() + 1.0F / 2.0F,
                                block.func_177230_c().func_185467_w().func_185847_b() * 0.8F);
                    }
                    onBlockMineFinishedDamagePlayerItem(player, x, y, z);

                    vanillaDestroyBlockProgressMap.remove(i);
                    partiallyMinedBlocksArray[i] = null;
                    if (curBlock.func_177958_n() == x && curBlock.func_177956_o() == y && curBlock.func_177952_p() == z)
                    {
                        curBlock = BlockPos.field_177992_a;
                    }
                    MultiMine.instance().debugPrint("Client wiped local finished block [" + x + "|" + y + "|" + z + "], at index " + i);
                }
                return;
            }
        }

        if (progress > 0.99)
        {
            MultiMine.instance().debugPrint("Client ignoring late arrival packet [" + x + "|" + y + "|" + z + "]");
            return;
        }

        if (freeIndex != -1)
        {
            partiallyMinedBlocksArray[freeIndex] = newBlock;
        }
        else
        {
            partiallyMinedBlocksArray[arrayOverWriteIndex++] = newBlock;
            if (arrayOverWriteIndex == partiallyMinedBlocksArray.length)
            {
                arrayOverWriteIndex = 0;
            }
        }
    }

    /**
     * Helper method to emulate vanilla behaviour of damaging your Item as you finish mining a Block.
     *
     * @param player Player doing the mining
     * @param x      Coordinates of the Block
     * @param y      Coordinates of the Block
     * @param z      Coordinates of the Block
     */
    private void onBlockMineFinishedDamagePlayerItem(EntityPlayer player, int x, int y, int z)
    {
        if (x != this.curBlock.func_177958_n() || y != curBlock.func_177956_o() || z != curBlock.func_177952_p())
        {
            return;
        }

        ItemStack itemStack = player.func_184614_ca();
        if (itemStack != null)
        {
            BlockPos pos = new BlockPos(x, y, z);
            itemStack.func_179548_a(player.field_70170_p, player.field_70170_p.func_180495_p(pos), pos, player);
            if (itemStack.field_77994_a == 0)
            {
                player.func_184611_a(EnumHand.MAIN_HAND, null);
            }
        }
    }

    /**
     * Called by the server via packet if there is more than the allowed amount of concurrent partial Blocks
     * in play. Causes the client to delete the corresponding local Block.
     */
    public void onServerSentPartialBlockDeleteCommand(BlockPos p)
    {
        for (int i = 0; i < partiallyMinedBlocksArray.length; i++)
        {
            if (partiallyMinedBlocksArray[i] != null && partiallyMinedBlocksArray[i].getPos().equals(p))
            {
                partiallyMinedBlocksArray[i] = null;
                vanillaDestroyBlockProgressMap.remove(i);
                break;
            }
        }
    }

    private void updateCloudTickReading()
    {
        // cache previous object
        DestroyBlockProgress dbp = vanillaDestroyBlockProgressMap.get(0);

        // execute code which gets the object assigned the private cloud tick value we want
        mc.field_71438_f.func_180441_b(0, new BlockPos((int) thePlayer.field_70165_t, (int) thePlayer.field_70163_u, (int) thePlayer.field_70161_v), 1);

        // read the needed value
        lastCloudTickReading = vanillaDestroyBlockProgressMap.get(0).func_82743_f();

        // execute code which destroys the helper object
        mc.field_71438_f.func_180441_b(0, new BlockPos((int) thePlayer.field_70165_t, (int) thePlayer.field_70163_u, (int) thePlayer.field_70161_v), 10);

        // if necessary restore previous object
        if (dbp != null)
        {
            vanillaDestroyBlockProgressMap.put(0, dbp);
        }
        // System.out.println("lastCloudTickReading is now "+lastCloudTickReading);
    }
}
